/* this file is copryright (C) 1996 Matthias Ettrich */

#include "forms.h"
#include <stdlib.h>
#include <stdio.h>
#include <dirent.h> 
#include <string.h>
#include <malloc.h>



/**** Forms and Objects ****/

typedef struct {
	FL_FORM *form_application;
	void *vdata;
	long ldata;
} FD_form_application;

FD_form_application * create_form_form_application(int, int);



/**** Callback routines ****/

void EditOkCB(FL_OBJECT *, long);
void EditApplyCB(FL_OBJECT *, long);
void EditCancelCB(FL_OBJECT *, long);
void EditBrowsePixmapCB(FL_OBJECT *, long);
void EditBrowseDirectoryCB(FL_OBJECT *, long);
void EditBrowseGroupFileCB(FL_OBJECT *, long);
void EditSetProgramCB(FL_OBJECT *, long);
void EditSetGroupCB(FL_OBJECT *, long);
void EditGroupCB(FL_OBJECT *, long);
void EditProgramCB(FL_OBJECT *, long);

void BrowseSelectCB(FL_OBJECT *, long);
void BrowseOkCB(FL_OBJECT *, long);
void BrowseCancelCB(FL_OBJECT *, long);


void button_cb(FL_OBJECT * ob, long data);
void remake(void);

const char* global_pattern;





/**** Forms and Objects ****/

typedef struct {
	FL_FORM *form_edit;
	FL_OBJECT *pixmap_preview;
	FL_OBJECT *input_label;
	FL_OBJECT *radio_program;
	FL_OBJECT *radio_group;
	FL_OBJECT *input_execute;
	FL_OBJECT *input_directory;
	FL_OBJECT *button_browse_pixmap;
	FL_OBJECT *input_pixmap;
	FL_OBJECT *button_browse_directory;
	FL_OBJECT *input_group_file;
	FL_OBJECT *button_browse_group_file;
	FL_OBJECT *button_ok;
	FL_OBJECT *button_apply;
	FL_OBJECT *button_cancel;
	void *vdata;
	long ldata;
} FD_form_edit;

FD_form_edit * create_form_form_edit(void);
typedef struct {
	FL_FORM *form_browse;
	FL_OBJECT *browser;
	FL_OBJECT *button_ok;
	FL_OBJECT *button_cancel;
	FL_OBJECT *label_title;
	void *vdata;
	long ldata;
} FD_form_browse;

void create_and_show_form_browse(const char* label, 
				 const char* pattern,
				 int options);





FD_form_application *fd_form_application;
FD_form_edit *fd_form_edit;
FD_form_browse *fd_form_browse;

#define newstr(x) strcpy((char*)malloc(strlen(x)+1),x)

#define GROUP_PIXMAP "apps"
#define NO_PIXMAP "question"

#define MAX_ITEMS 30


#define initial_width  440
#define initial_height  110
#define button_space 10

char* pixmap_path = PIXMAP_PATH;
char* panel_path = PANEL_PATH;
char* personal_path = PERSONAL_PATH;
char* real_personal_path = NULL; /* set in main */ 
char* home;


int number_of_items;
int the_edit_item;

typedef struct {
   char* label;
   char program; /* boolean */ 
   char* execute;
   char* directory;
   char* pixmap;
   char* group_file;
   FL_OBJECT* object;
} ITEM;


ITEM* items[MAX_ITEMS];

char* current_group = NULL;
char* current_group_label = NULL;
char group_changed = 0;
char* current_panel = NULL;


/* some prototypes */ 
char * full_path_name(const char* filename, const char* suffix);
char * full_pixmap_path_name(const char* filename);

void init_item_contents(int item);
void free_item_contents(int item);





void init_items(){
   number_of_items = 0;
}


char* newstr_without_suffix(const char* string){
   char* result = newstr(string);
   int i=strlen(result)-1;
   while(i && result[i]!='.')
     i--;
   if (i)
     result[i]='\0';
   return result;
}

char* newstr_with_suffix(const char* string, const char* suffix){
   char* result = strcpy((char*)malloc(strlen(string)+ strlen(suffix)+1),
			 string);
   strcat(result, suffix);
   return result;
}

FILE* open_file_for_writing(const char* name, const char* suffix){
   char* result;
   char* tmp;
   char* file_suffix;
   FILE * file;
   
   file_suffix = newstr_with_suffix(name, suffix);
   result = strcpy((char*)malloc(strlen(real_personal_path)+strlen(file_suffix) + 3),
		   real_personal_path);
   strcat(result, "/");
   strcat(result, file_suffix);
   file = fopen(result, "w");
   if (!file){
      /* maybe we should create the personal directory */ 
      tmp = strcpy((char*)malloc(strlen(real_personal_path)+ 12),
		   "mkdir ");
      strcat(tmp, real_personal_path);
      system(tmp);
      free(tmp);
      file = fopen(result, "w");
   }
   
   free(result);
   free(file_suffix);
   return file;
   
}


void read_group(char* name){
   char token[300];
   int i;
   FILE* file;
   char* save_name = newstr(name);
   
   for (i=0; i<number_of_items;i++){
      free_item_contents(i);
      free(items[i]);
   }
   number_of_items = 0;

   file = fopen(full_path_name(save_name,".grp"),"r");
   if (file){
      /* first scan the init line */ 
      fscanf(file, "%[^\n]", token);
      fscanf(file, "%1[\n]", token);token[0]='\0';
      while (!feof(file)){
	 items[number_of_items] = malloc(sizeof(ITEM));
	 init_item_contents(number_of_items);
	 
	 fscanf(file, "%[^\n]", token);
	 items[number_of_items]->program = !strcmp(token, "program");
	 fscanf(file, "%1[\n]", token);token[0]='\0';
	 
	 fscanf(file, "%[^\n]", token);
	 items[number_of_items]->label = newstr(token);
	 fscanf(file, "%1[\n]", token);token[0]='\0';

	 fscanf(file, "%[^\n]", token);
	 items[number_of_items]->execute = newstr(token);
	 fscanf(file, "%1[\n]", token);token[0]='\0';

	 fscanf(file, "%[^\n]", token);
	 items[number_of_items]->directory = newstr(token);
	 fscanf(file, "%1[\n]", token);token[0]='\0';
	 
	 fscanf(file, "%[^\n]", token);
	 items[number_of_items]->pixmap = newstr(token);
	 fscanf(file, "%1[\n]", token);token[0]='\0';
	 
	 fscanf(file, "%[^\n]", token);
	 items[number_of_items]->group_file = newstr(token);
	 fscanf(file, "%[\n]", token);token[0]='\0';
	 
	 number_of_items++;
      }
      fclose(file);
   }
   else
     {
	/* create a default group */ 
	items[number_of_items] = malloc(sizeof(ITEM));
	init_item_contents(number_of_items);
	free(items[number_of_items]->label);
	items[number_of_items]->label = newstr(current_group_label);
	free(items[number_of_items]->group_file);
	items[number_of_items]->group_file = newstr(current_group);
	items[number_of_items]->program = 0;
	
	number_of_items++;
	fl_show_alert("Xf-Panel 0.1 WARNING:","Group not found:", save_name, 1);
     }
   if (current_group)
     free(current_group);
   current_group = save_name;
   group_changed = 0;
   
   /* create objects */ 
   for (i=0; i<number_of_items;i++){
      items[i]->object = fl_create_pixmapbutton(FL_NORMAL_BUTTON,
						10,10,10,10, NULL);
      fl_add_object(fd_form_application->form_application, 
		    items[i]->object);
   }
}

void write_current_group(){
   FILE* file;
   int i;
   const char* tmp = fl_show_input("Current group name", current_group);
   
   if (tmp){
      free(current_group);
      current_group = newstr(tmp);
   }
   file = open_file_for_writing(current_group, ".grp");
   if (file){
      group_changed = 0;
      fprintf(file, "Group definition generated by Xf-Panel 0.1\n");
      for (i=0; i<number_of_items; i++){
	 if (items[i]->program)
	   fprintf(file, "program\n");
	 else
	   fprintf(file, "group\n");
	 fprintf(file, "%s\n%s\n%s\n%s\n%s\n", 
		 items[i]->label,
		 items[i]->execute,
		 items[i]->directory,
		 items[i]->pixmap,
		 items[i]->group_file);
      }
      
      fclose(file);
   }
}


void read_panel(char* name){
   
   int x, y, w, h;
   char token[300];
   FILE * file = fopen(full_path_name(name,".pan"),"r");
   if (file){
      
      /* first scan the init line */ 
      fscanf(file, "%[^\n]", token);
      fscanf(file, "%1[\n]", token);token[0]='\0';
      /* scan the geometry */ 
      fscanf(file, "%d %d %d %d", &x, &y, &w, &h);
      fscanf(file, "%1[\n]", token);token[0]='\0';
      /* scan the current group */ 
      fscanf(file, "%[^\n]", token);
      current_group = newstr(token);
      fclose(file);
            
      fd_form_application = create_form_form_application(w, h);
      fl_set_form_position(fd_form_application->form_application,
			   x, y);

      read_group(current_group);
      remake();
      
      fl_show_form(fd_form_application->form_application,
		   FL_PLACE_POSITION|FL_FREE_SIZE,
		   FL_FULLBORDER,
		   "Xf-Panel");
      if (current_panel)
	free(current_panel);
      current_panel = newstr(name);
   }
}

void write_current_panel(){
   FILE* file = open_file_for_writing(current_panel, ".pan");
   if (file){
      fprintf(file, "Panel definition generated by Xf-Panel 0.1\n");
      fprintf(file, "%d %d %d %d\n", 
	      fd_form_application->form_application->x,
	      fd_form_application->form_application->y,
	      fd_form_application->form_application->w,
	      fd_form_application->form_application->h);
      fprintf(file, "%s", current_group);
      fclose(file);
   }
}




void remake(void){
   int w, h;
   float button_width;
   char* pixmap_file;
   int i;
   w = fd_form_application->form_application->w;
   h = fd_form_application->form_application->h;
   
   /* calculate the width and space */
   button_width = (float)(w -  (number_of_items + 1) * button_space)
     /(float)number_of_items;
   
   for (i=0; i<number_of_items; i++){
/* newlib      fl_set_object_geometry((int)(i+1)*button_space + i * button_width,
			     2 * button_space,
			     button_width,
			     height-3*button_space) */ 

      items[i]->object->x = (int)(i+1)*button_space + i * button_width;
      items[i]->object->y = 1 * button_space;
      items[i]->object->w = button_width;
      items[i]->object->h = h-3*button_space;
     
      fl_set_object_label(items[i]->object, items[i]->label);
      
      fl_set_object_callback(items[i]->object, button_cb,i);

      if (items[i]->program){
	 pixmap_file = full_pixmap_path_name(items[i]->pixmap); 
	 if (pixmap_file){
	    fl_set_pixmapbutton_file(items[i]->object, pixmap_file);
	    free(pixmap_file);
	 } 
      }
      else {
	 pixmap_file = full_pixmap_path_name(GROUP_PIXMAP);
	 if (pixmap_file){
	    fl_set_pixmapbutton_file(items[i]->object, pixmap_file);
	    free(pixmap_file);
	 }
      }
      
      fl_set_object_resize(items[i]->object, FL_RESIZE_ALL);
      fl_set_object_gravity(items[i]->object, NorthGravity, SouthGravity);
      fl_set_object_callback(items[i]->object, button_cb,i);
   }
   
   
}




void free_item_contents(int item){
   /* free the item */ 
   free(items[item]->label);
   free(items[item]->execute);
   free(items[item]->directory);
   free(items[item]->pixmap);
   free(items[item]->group_file);
   if (items[item]->object)
     fl_delete_object(items[item]->object);
}

void init_item_contents(int item){
   items[item]->label = newstr("");
   items[item]->execute = newstr("");
   items[item]->directory = newstr("");
   items[item]->pixmap = newstr("");
   items[item]->group_file = newstr("");
   items[item]->program = 1;
   items[item]->object = NULL;
}



void set_edit_form_to_item(int item){

   the_edit_item = item;
   
   fl_set_input(fd_form_edit->input_label, items[item]->label);
   fl_set_input(fd_form_edit->input_execute, items[item]->execute);
   fl_set_input(fd_form_edit->input_directory, items[item]->directory);
   fl_set_input(fd_form_edit->input_pixmap, items[item]->pixmap);
   fl_set_input(fd_form_edit->input_group_file, items[item]->group_file);
   
   fl_set_button(fd_form_edit->radio_program,0);
   fl_set_button(fd_form_edit->radio_group,0);
   if (items[item]->program)
     EditSetProgramCB(0,0);
   else
     EditSetGroupCB(0,0);
}
		

void set_item_to_edit_form(int item){
   FL_OBJECT *obj = items[item]->object;
   items[item]->object = NULL;
   free_item_contents(item);
   items[item]->object = obj;
   
   items[item]->label = newstr(fl_get_input(fd_form_edit->input_label));
   items[item]->execute = newstr(fl_get_input(fd_form_edit->input_execute));
   items[item]->directory = newstr(fl_get_input(fd_form_edit->input_directory));
   items[item]->pixmap = newstr(fl_get_input(fd_form_edit->input_pixmap));
   items[item]->group_file = newstr(fl_get_input(fd_form_edit->input_group_file));

   items[item]->program = fl_get_button(fd_form_edit->radio_program);
   
   fl_freeze_form(fd_form_application->form_application);
   remake();
   fl_redraw_form(fd_form_application->form_application);
   fl_unfreeze_form(fd_form_application->form_application);
   
}
		


void edit_item(int item){
   
   set_edit_form_to_item(item);
   
   if (fd_form_edit->form_edit->visible)
     XRaiseWindow(fl_display, fd_form_edit->form_edit->window);
   else
     fl_show_form(fd_form_edit->form_edit,
		  FL_PLACE_CENTER | FL_FREE_SIZE, FL_FULLBORDER,
		  "Xf-Panel Edit");
}



void new_item(int item){
   int i;
   if (number_of_items < MAX_ITEMS){
      /* create and init the item */ 
      for (i=number_of_items; i>item+1;i--)
	items[i]=items[i-1];
      
      items[item + 1] = malloc(sizeof(ITEM));
      fl_freeze_form(fd_form_application->form_application);
      init_item_contents(item + 1);
      
      /* create new item in form */ 
      items[item+1]->object = fl_create_pixmapbutton(FL_NORMAL_BUTTON,
						     10,10,10,10, NULL);
      fl_add_object(fd_form_application->form_application, 
		    items[item+1]->object);
      
      number_of_items++;
      group_changed = 1;
      remake();
      fl_redraw_form(fd_form_application->form_application);
      fl_unfreeze_form(fd_form_application->form_application);
      edit_item(item+1);
   }
   
}

void delete_item(int item){
   int i;
   EditCancelCB(0,0);
   if (number_of_items>1){
      number_of_items--;
      fl_freeze_form(fd_form_application->form_application);
      free_item_contents(item);
      free(items[item]);
      for (i=item; i<number_of_items;i++)
	items[i]=items[i+1];
      group_changed = 1;
      remake();
      fl_redraw_form(fd_form_application->form_application);
      fl_unfreeze_form(fd_form_application->form_application);
   }
}




char * full_path_name(const char* filename, const char* suffix){
   /* TODO check for filename in all directories */ 
   char* result;
   FILE *file;
   char* file_suffix;
   if (!filename || !filename[0])
     return NULL;

   file_suffix = newstr_with_suffix(filename, suffix);
   
   /* first check $(HOME)/.Xf-Panel */ 
   result = strcpy((char*)malloc(strlen(real_personal_path)+strlen(file_suffix) + 3),
		   real_personal_path);
   strcat(result, "/");
   strcat(result, file_suffix);
   file = fopen(result, "r");
   if (file) {
      fclose(file);
      free(file_suffix);
      return result;
   }
   free(result);
   
   result = strcpy((char*)malloc(strlen(panel_path)+strlen(file_suffix)+2), panel_path);
   strcat(result, "/");
   strcat(result, file_suffix);
   file = fopen(result, "r");
   if (file){
      fclose(file);
      free(file_suffix);
      return result;
   }
   free(result);
   
   result = strcpy((char*)malloc(strlen(pixmap_path)+strlen(file_suffix)+2), pixmap_path);
   strcat(result, "/");
   strcat(result, file_suffix);
   file = fopen(result, "r");
   if (file){
      fclose(file);
      free(file_suffix);
      return result;
   }
   free(result);

   result = newstr(file_suffix);
   file = fopen(result, "r");
   if (file){
      fclose(file);
      return result;
   }
   free(result);
   free(file_suffix);
   
   return NULL;
}

char* full_pixmap_path_name(const char* file){
   char* tmp = full_path_name(file, ".xpm");
   if (!tmp)
     tmp = full_path_name(NO_PIXMAP, ".xpm"); 
   return tmp;
}


void button_cb(FL_OBJECT * ob, long data){
   int button = fl_get_button_numb(ob);
   if (button == 3){
      int Menu = fl_defpup(FL_ObjWin(ob),
			   "| Edit"
			   "| New"
			   "| Delete%l"
			   "| Save Group"
			   "| Save Panel"
			   );
      /*      fl_setpup_position(fd_form_application->form_application->x + ob->x,
			 fd_form_application->form_application->y + ob->y + 10);
			 */
      fl_setpup_title(Menu, current_group_label);
      switch (fl_dopup(Menu)) {
       case 1:
	 edit_item(data);
	 break;
       case 2:
	 new_item(data);
	 break;
       case 3:
	 delete_item(data);
	 break;
       case 4:
	 write_current_group();
	 break;
       case 5:
	   {
	      const char* tmp = fl_show_input("Panel Name", current_panel);
	      if (tmp){
		 free(current_panel);
		 current_panel = newstr(tmp);
		 write_current_panel();
	      }
	      
	   }
	 
	 break;
      }
      fl_freepup(Menu);
   }
   else {
      if (items[data]->program){
	 /* execute command */ 
	 char* tmp = (char*)malloc(strlen(items[data]->directory)+
				   strlen(items[data]->execute)+ 10);
	 if (items[data]->directory){
	    strcpy(tmp, "cd ");
	    strcat(tmp, items[data]->directory);
	    strcat(tmp, ";");
	    strcat(tmp, items[data]->execute);
	 }
	 else
	   strcpy(tmp, items[data]->execute);
	 strcat(tmp, "&"); 
	 system(tmp);
	 free(tmp);
      }
      else{
	 /* open other group */
	 char* tmp = newstr(items[data]->label);
	 if (group_changed){
	    if (fl_show_question("XF-Panel Warning: Current group contains unsaved changes.  Should I save it?",0))
	      write_current_group();
	 }
	 EditCancelCB(0,0);
	 fl_freeze_form(fd_form_application->form_application);
	 read_group(items[data]->group_file);
	 if (current_group_label)
	   free(current_group_label);
	 current_group_label = tmp;
	 remake();
	 fl_redraw_form(fd_form_application->form_application);
	 fl_unfreeze_form(fd_form_application->form_application);
      }
   }
}



/* callbacks for form form_edit */
void EditOkCB(FL_OBJECT *ob, long data)
{
   EditApplyCB(ob, data);
   EditCancelCB(ob, data);
}

void EditApplyCB(FL_OBJECT *ob, long data)
{
   group_changed = 1;
   set_item_to_edit_form(the_edit_item);
}

void EditCancelCB(FL_OBJECT *ob, long data)
{
   if (fd_form_edit->form_edit->visible)
     fl_hide_form(fd_form_edit->form_edit);
}


void EditBrowsePixmapCB(FL_OBJECT *ob, long data)
{
   EditSetProgramCB(ob, data);
   create_and_show_form_browse("Select Pixmap", ".xpm", 1);
}

void EditBrowseDirectoryCB(FL_OBJECT *ob, long data)
{
   EditSetProgramCB(ob, data);
   if (fl_show_fselector("Select Directory",
			 fl_get_input(fd_form_edit->input_directory),
			 "No pattern ;-)","do not remove this"))
     fl_set_input(fd_form_edit->input_directory, fl_get_directory());
}

void EditBrowseGroupFileCB(FL_OBJECT *ob, long data)
{
   EditSetGroupCB(ob, data);
   create_and_show_form_browse("Select Group", ".grp", 2);
}

void EditSetProgramCB(FL_OBJECT *ob, long data)
{
   if (!fl_get_button(fd_form_edit->radio_program)){
      fl_set_button(fd_form_edit->radio_program, 1);
      fl_set_button(fd_form_edit->radio_group, 0);
      EditProgramCB(ob, 1);
   }
   else
     if (data)
       EditProgramCB(ob, 0);
}

void EditSetGroupCB(FL_OBJECT *ob, long data)
{
   if (!fl_get_button(fd_form_edit->radio_group)){
      fl_set_button(fd_form_edit->radio_group, 1);
      fl_set_button(fd_form_edit->radio_program, 0);
      EditGroupCB(ob, data);
   }
}


void EditProgramCB(FL_OBJECT *ob, long data){
  static char* old = NULL;
  char* tmp = full_pixmap_path_name(fl_get_input(fd_form_edit->input_pixmap));
  if (tmp && (!old || strcmp(tmp, old) || data)){
    fl_set_pixmap_file(fd_form_edit->pixmap_preview, tmp);
    if (old)
      free(old);
    old = tmp;
      /* needed because of XForms bug */ 
     fl_redraw_form(fd_form_edit->form_edit); 
   }
}
   
void EditGroupCB(FL_OBJECT *ob, long data){
   char* tmp;
   tmp = full_pixmap_path_name(GROUP_PIXMAP);
   if (tmp){
      fl_set_pixmap_file(fd_form_edit->pixmap_preview, tmp);
      free(tmp);
      /* needed because of XForms bug */ 
      fl_redraw_form(fd_form_edit->form_edit); 
   }
}
   









/* callbacks for form form_browse */
void BrowseSelectCB(FL_OBJECT *ob, long data)
{
   char* tmp;
   if (data == 1) {  /* browse pixmap */
      tmp = full_pixmap_path_name(fl_get_browser_line(ob, fl_get_browser(ob)));
      if (tmp) { 
	 fl_set_pixmap_file(fd_form_edit->pixmap_preview, tmp);
	 free(tmp);
      }
   }
}

void BrowseOkCB(FL_OBJECT *ob, long data)
{
   if (data == 1){ /* browse pixmap */
      if (fl_get_browser(fd_form_browse->browser)){
	 fl_set_input(fd_form_edit->input_pixmap, 
		      fl_get_browser_line(fd_form_browse->browser,
					  fl_get_browser(fd_form_browse->browser)
					  )
		      );
      }
   }
   else{
      if (fl_get_browser(fd_form_browse->browser)){
	 fl_set_input(fd_form_edit->input_group_file, 
		      fl_get_browser_line(fd_form_browse->browser,
					  fl_get_browser(fd_form_browse->browser)
					  )
		      );
	 if (strlen(fl_get_input(fd_form_edit->input_label))==0){
	    fl_set_input(fd_form_edit->input_label,
			 fl_get_browser_line(fd_form_browse->browser,
					     fl_get_browser(fd_form_browse->browser)
					     ));
	 }
      }
   }
   
   
   BrowseCancelCB(ob, data);
}

void BrowseCancelCB(FL_OBJECT *ob, long data)
{
   if (data == 1) /* browse pixmap */ 
     EditProgramCB(ob, data);
   
   fl_hide_form(fd_form_browse->form_browse);
   fl_free_form(fd_form_browse->form_browse);
   fl_activate_all_forms();
   
}








/* Form definition file generated with fdesign. */

FD_form_application *create_form_form_application(int width, int height)
{
  FL_OBJECT *obj;
   FD_form_application *fdui = (FD_form_application *) fl_calloc(1, sizeof(FD_form_application));
   fdui->form_application = fl_bgn_form(FL_FLAT_BOX, width, height);
   obj = fl_add_box(FL_DOWN_BOX,2,2,width-4, height-4,"");
   fl_set_object_gravity(obj, NorthWestGravity, SouthEastGravity);
   fl_end_form();


  return fdui;
}
/*---------------------------------------*/



/* Form definition file generated with fdesign. */

FD_form_edit *create_form_form_edit(void)
{
  FL_OBJECT *obj;
  FD_form_edit *fdui = (FD_form_edit *) fl_calloc(1, sizeof(FD_form_edit));

  fdui->form_edit = fl_bgn_form(FL_NO_BOX, 480, 410);
  obj = fl_add_box(FL_UP_BOX,0,0,480,410,"");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
  fdui->input_label = obj = fl_add_input(FL_NORMAL_INPUT,90,410-320-30,220,30,"Label");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
  obj = fl_add_frame(FL_ENGRAVED_FRAME,10,410-160-140,460,140,"");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
  fdui->radio_program = obj = fl_add_checkbutton(FL_RADIO_BUTTON,20,410-280-30,90,30,"Program");
    fl_set_object_callback(obj, EditProgramCB, 1);
    fl_set_object_boxtype(obj,FL_FLAT_BOX);
    fl_set_object_lsize(obj,FL_MEDIUM_SIZE);
  obj = fl_add_frame(FL_ENGRAVED_FRAME,10,410-60-80,460,80,"");
  fdui->radio_group = obj = fl_add_checkbutton(FL_RADIO_BUTTON,20,410-120-30,80,30,"Group");
    fl_set_object_callback(obj, EditGroupCB, 0);
    fl_set_object_boxtype(obj,FL_FLAT_BOX);
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
  fdui->input_execute = obj = fl_add_input(FL_NORMAL_INPUT,90,410-250-30,300,30,"Execute");
    fl_set_object_callback(obj, EditSetProgramCB, 0);
    fl_set_input_return(obj, 1);
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
  fdui->input_directory = obj = fl_add_input(FL_NORMAL_INPUT,90,410-210-30,300,30,"Directory");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_callback(obj, EditSetProgramCB, 0);
    fl_set_input_return(obj, 1);
  fdui->pixmap_preview = obj = fl_add_pixmap(FL_NORMAL_PIXMAP,350,410-310-90,100,90,"");
  obj = fl_add_text(FL_NORMAL_TEXT,20,410-360-40,150,40,"Xf-Panel 0.1");
    fl_set_object_lsize(obj,FL_HUGE_SIZE);
    fl_set_object_lalign(obj,FL_ALIGN_CENTER);
    fl_set_object_lstyle(obj,FL_ITALIC_STYLE+FL_ENGRAVED_STYLE);
  obj = fl_add_text(FL_NORMAL_TEXT,170,410-360-30,140,30,"(c) 1996 Matthias Ettrich");
    fl_set_object_lalign(obj,FL_ALIGN_CENTER);
  fdui->button_browse_pixmap = obj = fl_add_button(FL_NORMAL_BUTTON,400,410-170-30,60,30,"Browse...");
    fl_set_object_callback(obj,EditBrowsePixmapCB,0);
  fdui->input_pixmap = obj = fl_add_input(FL_NORMAL_INPUT,90,410-170-30,300,30,"Pixmap");
    fl_set_object_callback(obj, EditSetProgramCB, 1);
    fl_set_input_return(obj, 1);
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
  fdui->button_browse_directory = obj = fl_add_button(FL_NORMAL_BUTTON,400,410-210-30,60,30,"Browse...");
    fl_set_object_callback(obj,EditBrowseDirectoryCB,0);
  fdui->input_group_file = obj = fl_add_input(FL_NORMAL_INPUT,90,410-80-30,300,30,"Group File");
    fl_set_object_callback(obj, EditSetGroupCB, 0);
    fl_set_input_return(obj, 1);
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
  fdui->button_browse_group_file = obj = fl_add_button(FL_NORMAL_BUTTON,400,410-80-30,60,30,"Browse...");
    fl_set_object_callback(obj,EditBrowseGroupFileCB,0);
  fdui->button_ok = obj = fl_add_button(FL_RETURN_BUTTON,10,410-10-30,140,30,"OK");
    fl_set_object_lsize(obj,FL_MEDIUM_SIZE);
    fl_set_object_callback(obj,EditOkCB,0);
  fdui->button_apply = obj = fl_add_button(FL_NORMAL_BUTTON,170,410-10-30,140,30,"Apply");
    fl_set_object_lsize(obj,FL_MEDIUM_SIZE);
    fl_set_object_callback(obj,EditApplyCB,0);
  fdui->button_cancel = obj = fl_add_button(FL_NORMAL_BUTTON,330,410-10-30,140,30,"Cancel");
    fl_set_object_lsize(obj,FL_MEDIUM_SIZE);
    fl_set_object_callback(obj,EditCancelCB,0);
  fl_end_form();
   
  return fdui;
}
/*---------------------------------------*/

const char* global_pattern;

int my_select(const struct dirent *entry){
   int n;
   for (n=0; entry->d_name[n]; n++);
   for (;n>0&&entry->d_name[n]!='.';n--);
   return !strcmp(global_pattern, entry->d_name + n);
}


void create_and_show_form_browse(const char* label, 
				 const char* pattern,
				 int options)
{
   /* struct dirent **namelist; */ 
   struct dirent *de;
   DIR* mydir;
   /* int i,n; */ 
   char* tmp;
   FL_OBJECT *obj;
   FD_form_browse *fdui = (FD_form_browse *) fl_calloc(1, sizeof(FD_form_browse));
   
   fl_deactivate_all_forms();
   
   fdui->form_browse = fl_bgn_form(FL_NO_BOX, 190, 320);
   obj = fl_add_box(FL_UP_BOX,0,0,190,320,"");
   fdui->browser = obj = fl_add_browser(FL_HOLD_BROWSER,10,320-50-220,170,220,"");
   fl_set_object_callback(obj,BrowseSelectCB,options);
   fdui->button_ok = obj = fl_add_button(FL_RETURN_BUTTON,10,320-10-30,80,30,"OK");
   fl_set_object_lsize(obj,FL_MEDIUM_SIZE);
   fl_set_object_callback(obj,BrowseOkCB,options);
   fdui->button_cancel = obj = fl_add_button(FL_NORMAL_BUTTON,100,320-10-30,80,30,"Cancel");
   fl_set_object_lsize(obj,FL_MEDIUM_SIZE);
   fl_set_object_callback(obj,BrowseCancelCB,options);
   fdui->label_title = obj = fl_add_text(FL_NORMAL_TEXT,10,320-280-20,170,20,label);
   fl_set_object_lsize(obj,FL_MEDIUM_SIZE);
   fl_end_form();
   
   global_pattern = pattern;
   /*   
   n = scandir(real_personal_path, &namelist, my_select, alphasort);
   for (i=0; i<n; i++)
     {
	fl_add_browser_line(fdui->browser, namelist[i]->d_name);
     }
   n = scandir(panel_path, &namelist, my_select, alphasort);
   for (i=0; i<n; i++)
     {
	fl_add_browser_line(fdui->browser, namelist[i]->d_name);
     }
   n = scandir(pixmap_path, &namelist, my_select, alphasort);
   for (i=0; i<n; i++)
     {
	fl_add_browser_line(fdui->browser, namelist[i]->d_name);
     }
   n = scandir(".", &namelist, my_select, alphasort);
   for (i=0; i<n; i++)
     {
	fl_add_browser_line(fdui->browser, namelist[i]->d_name);
     }
   
     */
   
   
   /* scandir is obsolete :-) So this should work on more than linux :-) */ 
   
   mydir = opendir(real_personal_path);
   if (mydir){
      do {
	 de = readdir(mydir);
	 if (de && my_select(de)){
	    tmp = newstr_without_suffix(de->d_name);
	    fl_add_browser_line(fdui->browser, tmp);
	    free(tmp);
	 }
      } while (de);
      closedir(mydir);
   }
   mydir = opendir(panel_path);
   if (mydir){
      do {
	 de = readdir(mydir);
	 if (de && my_select(de)){
	    tmp = newstr_without_suffix(de->d_name);
	    fl_add_browser_line(fdui->browser, tmp);
	    free(tmp);
	 }
      }
      while (de);
      closedir(mydir);
   }
   mydir = opendir(pixmap_path);
   if (mydir){
      do {
	 de = readdir(mydir);
	 if (de && my_select(de)){
	    tmp = newstr_without_suffix(de->d_name);
	    fl_add_browser_line(fdui->browser, tmp);
	    free(tmp);
	 }
      } while (de);
      closedir(mydir);
   }
   
   /* here the browser should be sorted. */ 
   
   
   
   fd_form_browse = fdui;
   
   fl_show_form(fd_form_browse->form_browse,FL_PLACE_CENTER|FL_FREE_SIZE,FL_FULLBORDER,
		"Xf-Panel Browser");
   
}
/*---------------------------------------*/





int main(int argc, char *argv[])
{
   char* start_panel = "default";
   char* tmp;
   fl_set_border_width(-2);
   
   fl_initialize(&argc, argv, "Xf-Panel", 0, 0);
   fl_setpup_fontstyle(FL_NORMAL_STYLE);
   fl_setpup_fontsize(FL_NORMAL_SIZE);

   

   home = getenv("HOME");
   if (!home){
      fprintf(stderr, "Xf-Panel FATAL ERROR: $HOME not defined\n");
   }
   real_personal_path = strcpy((char*)malloc(strlen(home)+strlen(personal_path)+ 3),
			       home);
   strcat(real_personal_path, "/");
   strcat(real_personal_path, personal_path);

   tmp = getenv("XF_PANEL_PATH");
   if (tmp)
     panel_path = tmp;
   tmp = getenv("XF_PANEL_PIXMAP_PATH");
   if (tmp)
     pixmap_path = tmp;
   
   if (argc>1)
     start_panel = argv[1];
   

   if(!full_path_name("Xf-Panel", ".README")){
      fprintf(stderr, "Xf-Panel FATAL ERROR: Cannot locate global configuration files\n\n");
      fprintf(stderr, "It seems that Xf-Panel isn't correctly installed.\n");
      fprintf(stderr, "Please read the Installation-Guide.\n\n");
      exit(1);
   }
   
   

   
   
   init_items();
   
   current_group_label = newstr("Main");
   
   read_panel(start_panel);
   if(!current_panel){
      fprintf(stderr, "Xf-Panel FATAL ERROR: Cannot open panel %s\n\n", start_panel);
      exit(1);
   }
   
   fd_form_edit = create_form_form_edit();
   fl_set_form_dblbuffer(fd_form_edit->form_edit, 1);
   
   /* fill-in form initialization code */

   /* show the first form */
   fl_show_form(fd_form_application->form_application,FL_PLACE_CENTER|FL_FREE_SIZE,FL_FULLBORDER,
		"Xf-Panel");

   while(1)
     fl_do_forms();
   return 0;
}
